import contextlib

from PyQt5.QtWidgets import *
from PyQt5.QtCore import Qt, QThreadPool, pyqtSignal, QThread

from .worker import Worker
from .. import Pyscandl
from ..modules.autodl import Controller
from ..modules.fetchers import FetcherEnum


class QStdoutText(QTextEdit):
    _write = pyqtSignal(object)

    def __init__(self):
        super().__init__()
        self.setReadOnly(True)
        self._write.connect(self.new_text)

    def write(self, txt):
        self._write.emit(txt)

    def new_text(self, txt):
        bottom = self.verticalScrollBar().value() == self.verticalScrollBar().maximum()
        self.insertPlainText(str(txt))
        if bottom:
            self.verticalScrollBar().setValue(self.verticalScrollBar().maximum())

    def flush(self):
        pass



class MainWindow(QMainWindow):
    _autodl_progress_count = pyqtSignal(int)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._threadpool = QThreadPool()
        self._controller = Controller()
        self.setupUi()
        self.show()

    def setupUi(self):
        self.setWindowTitle("pyscandl")
        tabs = QTabWidget()
        manga_panel = self.manga_panel()
        autodl_panel = self.autodl_panel()
        manual_panel = self.manual_panel()
        tabs.addTab(manga_panel, "&manga")
        tabs.addTab(autodl_panel, "&autodl")
        tabs.addTab(manual_panel, "ma&nual")
        self.setCentralWidget(tabs)

    def manga_panel(self):
        panel = QWidget()
        layout = QVBoxLayout()
        top_layout = QHBoxLayout()

        # setting up the list of mangas
        self._mangas = QComboBox()
        self._mangas.addItems(self._controller.list_mangas(all=True, format=False))
        current = self._controller.manga_info(self._mangas.currentText())
        top_layout.addWidget(self._mangas)
        self._mangas.currentTextChanged.connect(self._update_info)

        # setting up the control pane
        see_chaps_button = QPushButton("see all chapters")
        see_chaps_button.clicked.connect(self._chapter_list_message)
        new_button = QPushButton("new")
        new_button.clicked.connect(self._create_dialog)
        edit_button = QPushButton("edit")
        edit_button.clicked.connect(self._edit_dialog)
        delete_button = QPushButton("delete")
        delete_button.clicked.connect(self._delete_confirm)
        top_layout.addWidget(see_chaps_button)
        top_layout.addWidget(new_button)
        top_layout.addWidget(edit_button)
        top_layout.addWidget(delete_button)

        # setting up the info pane
        bottom = QGroupBox()
        bottom.setTitle("info")
        self._form = QFormLayout()
        bottom.setLayout(self._form)
        self._form.addRow("- link:", QLabel(current[2]))
        self._form.addRow("- fetcher:", QLabel(current[1]))
        self._form.addRow("- archived:", QLabel("YES" if current[3] else "NO"))
        if current[4]:
            self._form.addRow("- number of chapters downloaded:", QLabel(str(len(current[4]))))
            self._form.addRow("- last chapter:", QLabel(str(current[4][0])))
        else:
            self._form.addRow("- number of chapters downloaded:", QLabel("no chapters downloaded yet"))
            self._form.addRow("- last chapter:", QLabel("no chapters downloaded yet"))


        # adding everything to the pannel
        layout.addLayout(top_layout)
        layout.addWidget(bottom)
        panel.setLayout(layout)
        return panel

    def autodl_panel(self):
        panel = QWidget()
        layout = QVBoxLayout()

        # constructing the form
        param_line = QHBoxLayout()
        param_line.addLayout(self._folder_widget("Save location:"))
        quiet = QCheckBox("quiet")
        param_line.addWidget(quiet)
        tiny = QCheckBox("tiny")
        tiny.setToolTip("don't write the manga name in the title (useful if using ebook libraries)")
        param_line.addWidget(tiny)
        layout.addLayout(param_line)
        mode = QComboBox()
        mode.addItem("PDF", (True, False, False))
        mode.addItem("image", (False, False, True))
        mode.addItem("keep", (False, True, False))
        mode.setToolTip("""- pdf: downloads only the pdf of the manga
- keep: downloads the pdf but also keep the images in a chapter subfolder
- image: downloads only the images in a chapter subfolder and don't create the pdf""")
        param_line.addWidget(mode)

        # progress bar
        self._autodl_progress = QProgressBar()
        self._autodl_progress_count.connect(self._autodl_progress.setValue)
        layout.addWidget(self._autodl_progress)

        # output
        self._autodl_out = QStdoutText()
        layout.addWidget(self._autodl_out)

        # launching actions
        buttons = QHBoxLayout()
        buttons.setAlignment(Qt.AlignRight)
        download = QPushButton("Download")
        download.clicked.connect(self._worker_autodl_download)
        download.clicked.connect(self._autodl_out.clear)
        buttons.addWidget(download)
        scan = QPushButton("Scan")
        scan.setToolTip("checks if there is any new chapters but doesn't download them")
        scan.clicked.connect(self._worker_autodl_scan)
        scan.clicked.connect(self._autodl_out.clear)
        buttons.addWidget(scan)
        layout.addLayout(buttons)

        panel.setLayout(layout)
        return panel

    def _autodl_download(self, sender):
        parent = sender.parent()
        pdf, keep, image = parent.findChild(QComboBox).currentData()
        output = parent.findChild(QLineEdit).text()
        quiet, tiny = [bool(check.checkState()) for check in parent.findChildren(QCheckBox)]
        controller = Controller(
            output=output if output else ".",
            quiet=quiet,
            tiny=tiny
        )
        mangas = controller.list_mangas(format=False)
        self._autodl_progress.setMaximum(len(mangas))
        with contextlib.redirect_stdout(self._autodl_out):
            for manga in mangas:
                self._autodl_progress_count.emit(self._autodl_progress.value() + 1)
                controller.scan(manga)
                controller.download(manga, pdf, keep, image)
            controller.save()
            self._autodl_progress_count.emit(self._autodl_progress.value() + 1)
        if not quiet:
            self._autodl_out.write(f"{controller.downloads} chapters downloaded")

    def _worker_autodl_download(self):
        self._autodl_progress.reset()
        buttons = self.sender().parent().findChildren(QPushButton)[1:]
        buttons[0].setDisabled(True)
        buttons[1].setDisabled(True)
        print("starting dl")
        worker = Worker(self._autodl_download, self.sender())
        worker.signals.result.connect(lambda x: print(x))
        worker.signals.finished.connect(lambda: print("finished"))
        worker.signals.finished.connect(buttons[0].setEnabled)
        worker.signals.finished.connect(buttons[1].setEnabled)

        self._threadpool.start(worker)


    def _autodl_scan(self):
        controller = Controller()
        mangas = controller.list_mangas(format=False)
        self._autodl_progress.setMaximum(len(mangas))
        with contextlib.redirect_stdout(self._autodl_out):
            for manga in mangas:
                self._autodl_progress_count.emit(self._autodl_progress.value() + 1)
                controller.scan(manga)
            self._autodl_progress_count.emit(self._autodl_progress.value() + 1)

    def _worker_autodl_scan(self):
        self._autodl_progress.reset()
        buttons = self.sender().parent().findChildren(QPushButton)[1:]
        buttons[0].setDisabled(True)
        buttons[1].setDisabled(True)
        print("starting dl")
        worker = Worker(self._autodl_scan)
        worker.signals.result.connect(lambda x: print(x))
        worker.signals.finished.connect(lambda: print("finished"))
        worker.signals.finished.connect(buttons[0].setEnabled)
        worker.signals.finished.connect(buttons[1].setEnabled)

        self._threadpool.start(worker)

    def manual_panel(self):
        panel = QWidget()
        layout = QVBoxLayout()
        top_layout = QGridLayout()

        # constructing 2 line form in the QGridLayout
        ## first line: save location, quiet, tiny, download mode
        first_line = QHBoxLayout()
        first_line.addLayout(self._folder_widget("Save location:"))
        quiet = QCheckBox("quiet")
        first_line.addWidget(quiet)
        tiny = QCheckBox("tiny")
        tiny.setToolTip("don't write the manga name in the title (useful if using ebook libraries)")
        first_line.addWidget(tiny)
        mode = QComboBox()
        mode.addItem("PDF", (True, False, False))
        mode.addItem("image", (False, False, True))
        mode.addItem("keep", (False, True, False))
        mode.setToolTip("""- pdf: downloads only the pdf of the manga
- keep: downloads the pdf but also keep the images in a chapter subfolder
- image: downloads only the images in a chapter subfolder and don't create the pdf""")
        first_line.addWidget(mode)

        ## second line: fetcher, link, starting point, end mode, end point
        second_line = QHBoxLayout()
        second_line.addWidget(QLabel("Fetcher:"))
        fetchers = QComboBox()
        for fetcher in FetcherEnum.list():
            fetchers.addItem(fetcher, FetcherEnum.get(fetcher))
        second_line.addWidget(fetchers)
        second_line.addWidget(QLabel("Link:"))
        second_line.addWidget(QLineEdit())
        all = QCheckBox("all")
        second_line.addWidget(all)
        second_line.addWidget(QLabel("Start:"))
        start = QDoubleSpinBox()
        start.setMaximum(2147483647)
        start.setValue(1)
        second_line.addWidget(start)
        second_line.addWidget(QLabel("Stop:"))
        end_mode = QComboBox()
        end_mode.addItem("count chapters", True)
        end_mode.addItem("last chapter", False)
        end_mode.setToolTip("""- count chapters: once the total number of chapters downloaded equals the specified number
        - last chapter: once the chapter downloaded equals or exceeds the the specifed number""")
        second_line.addWidget(end_mode)
        stop = QDoubleSpinBox()
        stop.setMaximum(2147483647)
        stop.setValue(1)
        second_line.addWidget(stop)
        all.stateChanged.connect(end_mode.setDisabled)
        all.stateChanged.connect(stop.setDisabled)

        # output
        self._manual_out = QStdoutText()

        # launching actions
        buttons = QHBoxLayout()
        buttons.setAlignment(Qt.AlignRight)
        download = QPushButton("Download")
        download.clicked.connect(self._worker_manual_download)
        buttons.addWidget(download)
        cancel = QPushButton("Cancel")
        buttons.addWidget(cancel)

        # building the layout
        layout.addLayout(first_line)
        layout.addLayout(second_line)
        layout.addWidget(self._manual_out)
        layout.addLayout(buttons)
        panel.setLayout(layout)
        return panel

    def _manual_download(self, sender):
        parent = sender.parent()
        quiet, tiny, all = [bool(check.checkState()) for check in parent.findChildren(QCheckBox)]
        output, link = [line_ed.text() for line_ed in parent.findChildren(QLineEdit)[:2]]
        start, stop = [int(spin.value()) if spin.value() == int(spin.value()) else spin.value() for spin in parent.findChildren(QDoubleSpinBox)]
        dl_mode, fetcher, stop_mode = [combo.currentData() for combo in parent.findChildren(QComboBox)]
        manual_dl = Pyscandl(
            fetcher=fetcher,
            chapstart=start,
            output=output if output != "" else ".",
            link=link,
            all=all,
            pdf=dl_mode[0],
            keep=dl_mode[1],
            image=dl_mode[2],
            download_number=stop if stop_mode else 1,
            chapend=stop if not stop_mode else 0,
            quiet=quiet,
            tiny=tiny
        )

        with contextlib.redirect_stdout(self._manual_out):
            manual_dl.full_download()

    def _worker_manual_download(self):
        self.sender().setDisabled(True)
        print("starting dl")
        worker = Worker(self._manual_download, self.sender())
        worker.signals.result.connect(lambda x: print(x))
        worker.signals.finished.connect(lambda: print("finished"))
        worker.signals.finished.connect(self.sender().setEnabled)

        self._threadpool.start(worker)

    def _folder_widget(self, label=""):
        folder_widget = QHBoxLayout()
        if label:
            folder_widget.addWidget(QLabel(label))
        path = QLineEdit()
        file_button = QPushButton(QFileIconProvider().icon(QFileIconProvider.Folder), "")
        file_button.clicked.connect(self._folder_select)
        folder_widget.addWidget(path)
        folder_widget.addWidget(file_button)
        return folder_widget

    def _folder_select(self):
        sender = self.sender()
        folder_select = QFileDialog()
        folder_select.setFileMode(QFileDialog.DirectoryOnly)
        folder_select.fileSelected.connect(sender.parent().findChild(QLineEdit).setText)
        folder_select.exec()

    def _chapter_list_message(self):
        # setting up the chapterlist message box
        chaplist = QMessageBox()
        chaplist.setIcon(QMessageBox.Information)
        chaplist.setWindowTitle("chapter list")
        chaplist.setText(f"chapter list for {self._mangas.currentText()}:")
        chaplist.setInformativeText(" ".join([str(chap) for chap in self._controller.manga_info(self._mangas.currentText())[4]]))
        chaplist.exec()

    def _delete_confirm(self):
        delete_msg = QMessageBox.question(self, "deleting manga", f"are you sure you want to delete {self._mangas.currentText()}", QMessageBox.Yes, QMessageBox.No)

        if delete_msg == QMessageBox.Yes:
            self._controller.delete_manga(self._mangas.currentText())
            self._controller.save()
            self._mangas.removeItem(self._mangas.currentIndex())

    def _edit_dialog(self):
        create = QDialog()
        layout = QVBoxLayout()
        form = QFormLayout()
        manga = self._controller.manga_info(self._mangas.currentText())

        # info form
        link = QLineEdit()
        link.setText(manga[2])
        form.addRow("Link:", link)
        fetchers = QComboBox()
        fetchers.addItems(FetcherEnum.list(standalone=False))
        fetchers.setCurrentIndex(fetchers.findText(manga[1]))
        form.addRow("Fetcher:", fetchers)
        mode = QComboBox()
        mode.addItems(["set", "add", "delete"])
        mode.setToolTip("""- set: set the listed chapters as the chapters stored
- add: add the listed chapters to the chapters stored
- delete: delete the listed chapters from the chapters stored""")
        form.addRow("Chapter mode:", mode)
        chaps = QTextEdit()
        chaps.setText(" ".join([str(chap) for chap in manga[4]]))
        form.addRow("Chapters:", chaps)
        archived = QCheckBox()
        archived.setChecked(manga[3])
        form.addRow("Archived ?", archived)

        # accept and cancel buttons
        buttons = QDialogButtonBox(QDialogButtonBox.Save | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self._edit_save)
        buttons.accepted.connect(create.accept)
        buttons.rejected.connect(create.reject)

        layout.addLayout(form)
        layout.addWidget(buttons)
        create.setLayout(layout)
        create.exec()

    def _edit_save(self):
        sender = self.sender()
        form = sender.parent().layout().itemAt(0)
        link = form.itemAt(1).widget().text()
        if not link:
            self._error_popups(
                icon=QMessageBox.Critical,
                title="name error",
                text="an error when entering the link was detected:",
                info_text="please enter a link for the new manga"
            )
            sender.parent().reject()
        else:
            fetcher = form.itemAt(3).widget().currentText()
            archived = form.itemAt(9).widget().isChecked()
            try:
                chaps = [float(chap) if "." in chap else int(chap) for chap in
                         form.itemAt(7).widget().toPlainText().replace("\n", " ").split(" ") if chap != ""]

                mode = form.itemAt(5).widget().currentText()
                if mode == "set":
                    old_chaps = self._controller.manga_info(self._mangas.currentText())[4]
                    self._controller.rm_chaps(self._mangas.currentText(), [str(chap) for chap in old_chaps])
                    self._controller.edit(self._mangas.currentText(), link, fetcher, chaps, archived)
                elif mode == "add":
                    self._controller.edit(self._mangas.currentText(), link, fetcher, chaps, archived)
                elif mode == "delete":
                    self._controller.rm_chaps(self._mangas.currentText(), [str(chap) for chap in chaps])
                    self._controller.edit(self._mangas.currentText(), link, fetcher, None, archived)
                self._controller.save()
                self._update_info()
            except ValueError as e:
                # setting up the error message box
                self._error_popups(
                    icon=QMessageBox.Critical,
                    title="chapter list error",
                    text="an error when entering the chapters was detected:",
                    info_text="please only type chapter numbers like 23, 0 or 2.3 separated only by spaces or carriage returns",
                    detailed_text=str(e))
                sender.parent().reject()

    def _create_dialog(self):
        create = QDialog()
        layout = QVBoxLayout()
        form = QFormLayout()

        # info form
        form.addRow("Name:", QLineEdit())
        form.addRow("Link:", QLineEdit())
        fetchers = QComboBox()
        fetchers.addItems(FetcherEnum.list(standalone=False))
        form.addRow("Fetcher:", fetchers)
        form.addRow("Chapters (optional):", QTextEdit())
        form.addRow("Archived ?", QCheckBox())

        # accept and cancel buttons
        buttons = QDialogButtonBox(QDialogButtonBox.Save | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self._create_save)
        buttons.accepted.connect(create.accept)
        buttons.rejected.connect(create.reject)

        layout.addLayout(form)
        layout.addWidget(buttons)
        create.setLayout(layout)
        create.exec()

    def _create_save(self):
        sender = self.sender()
        form = sender.parent().layout().itemAt(0)
        name = form.itemAt(1).widget().text()
        link = form.itemAt(3).widget().text()
        if not name:
            self._error_popups(
                icon=QMessageBox.Critical,
                title="name error",
                text="an error when entering the name was detected:",
                info_text="please enter a name for the new manga"
            )
            sender.parent().reject()
        elif not link:
            self._error_popups(
                icon=QMessageBox.Critical,
                title="name error",
                text="an error when entering the link was detected:",
                info_text="please enter a link for the new manga"
            )
            sender.parent().reject()
        else:
            fetcher = form.itemAt(5).widget().currentText()
            archived = form.itemAt(9).widget().isChecked()
            try:
                chaps = [float(chap) if "." in chap else int(chap) for chap in form.itemAt(7).widget().toPlainText().replace("\n", " ").split(" ") if chap != ""]

                self._controller.add(name, link, fetcher, chaps, archived)
                self._controller.save()
                self._mangas.addItem(name)
            except ValueError as e:
                # setting up the error message box
                self._error_popups(
                    icon=QMessageBox.Critical,
                    title="chapter list error",
                    text="an error when entering the chapters was detected:",
                    info_text="please only type chapter numbers like 23, 0 or 2.3 separated only by spaces or carriage returns",
                    detailed_text=str(e))
                sender.parent().reject()

    def _update_info(self):
        current = self._controller.manga_info(self._mangas.currentText())
        self._form.itemAt(1).widget().setText(current[2])
        self._form.itemAt(3).widget().setText(current[1])
        self._form.itemAt(5).widget().setText("YES" if current[3] else "NO")
        if current[4]:
            self._form.itemAt(7).widget().setText(str(len(current[4])))
            self._form.itemAt(9).widget().setText(str(current[4][0]))
        else:
            self._form.itemAt(7).widget().setText("no chapters downloaded yet")
            self._form.itemAt(9).widget().setText("no chapters downloaded yet")

    def _error_popups(self, icon=QMessageBox.Critical, title="error", text=None, info_text=None, detailed_text=None):
        error = QMessageBox()
        error.setIcon(icon)
        error.setWindowTitle(title)
        if text:
            error.setText(text)
        if info_text:
            error.setInformativeText(info_text)
        if detailed_text:
            error.setDetailedText(detailed_text)
        error.exec()



import contextlib

from .excepts import DryNoSauceHere, TooManySauce, EmptyChapter, DelayedRelease
from .fetchers.fetcher import StandaloneFetcher
from PIL import Image
import requests
import os
from re import sub as re_sub
from sys import stderr
from wand.image import Image


class Pyscandl:
	"""
	The main object of the program. It is responsible of the downloads and controls the fetchers for you.
	"""
	def __init__(self, fetcher, chapstart=1, output:str=".", pdf:bool=True, keep:bool=False, image:bool=False, all:bool=False, link:str=None, manga:str=None, download_number:int=1, chapend=0, quiet:bool=False, skip:int=0, tiny:bool=False):
		"""
		Initialize this instance of the pyscandl downloader, it needs either manga or link to work.

		:param fetcher: fetcher object related to the download
		:param chapstart: first chapter to be downloaded
		:type chapstart: int/float/str
		:param output: output folder
		:type output: str
		:param pdf: tell if the result should be kept as a pdf
		:type pdf: bool
		:param keep: tell if the result should be kept as a pdf and as a collection of images
		:type keep: bool
		:param image: tell if the result should be kept as a collection of images
		:type image: bool
		:param all: download all the chapters that are available after chapstart
		:type all: bool
		:param link: link of the manga to download
		:type link: str
		:param manga: identification tag of the manga *(see every fetcher for their different variations)*
		:type manga: str
		:param download_number: number of chapters to download
		:type download_number: int
		:param chapend: chapter to end the download on, if non exstant the download will stop once the next to download chapter number is greater than it
		:type chapend: int/float/str
		:param quiet: should the program not output any information about what it is doing in the console
		:type quiet: bool
		:param skip: number of images to skip on the first chapter being downloaded *(useful if running in image mode)*
		:type skip: int
		:param tiny: should the name of every downloaded scan be minified and only include the chapter number and the chapter title
		:type tiny: bool

		:raises DryNoSauceHere: neither link or manga was specified
		:raises TooManySauce: both link and manga were specified
		"""

		if link is not None and manga is None or link is None and manga is not None:
			if issubclass(fetcher, StandaloneFetcher):
				self.fetcher = fetcher(link=link, manga=manga)
			else:
				self.fetcher = fetcher(link=link, manga=manga, chapstart=chapstart)
		elif link is None and manga is None:
			raise DryNoSauceHere
		else:
			raise TooManySauce

		# in case windows is the os, remove the banned characters
		if os.name == "nt":
			manga_name = re_sub(r'[\\/*?:"<>|]', u"█", self.fetcher.manga_name)
		else:
			manga_name = self.fetcher.manga_name

		# creating output folder
		self._output = (output[-1] == "/" and output or output + "/") + manga_name + "/"

		if not os.path.exists(self._output):
			os.makedirs(self._output)

		self._header = {"User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/77.0.3865.120 Safari/537.36",
					   "Set-Cookie": f"domain={self.fetcher.domain}"}
		self._header.update(self.fetcher.headers)
		self._nskip = skip
		self._quiet = quiet

		# select download mode
		self._pdf = pdf
		self._keep = keep
		self._image = image

		self._all = all
		self._download_number = download_number
		self._chapend = float(chapend) if "." in str(chapend) else int(chapend)
		self._path = f"{self._output}ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}/"  # save path for images
		self._img_bin_list = []
		self._tiny = tiny

		# in case windows is the os, remove the banned characters
		if os.name == "nt":
			chapter_name = re_sub(r'[\\/*?:"<>|]', u"█", self.fetcher.chapter_name)
		else:
			chapter_name = self.fetcher.chapter_name

		if self._tiny:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"
		else:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{manga_name} - {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - {self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}{manga_name} - ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"

		self._banlist = []
		ban_path = f"{os.path.dirname(os.path.abspath(__file__))}/../banlist"
		for img in os.listdir(ban_path):
			with open(f"{ban_path}/{img}", "rb") as img_bin:
				self._banlist.append(img_bin.read())

	def _dl_image(self):
		"""
		Downloads the currently selected image.
		"""

		if not os.path.exists(self._path):
			os.makedirs(self._path)

		with open(f"{self._path}{self.fetcher.npage}.{self.fetcher.ext}", "wb") as img:
			img.write(requests.get(self.fetcher.image, headers=self._header).content)
			if not self._quiet:
				print(".", end="", flush=True)

	def full_chapter(self):
		"""
		Fetching all the images of the chapter and storing them in RAM.
		"""

		if not self._quiet:
			if isinstance(self.fetcher, StandaloneFetcher):
				print(f"fetching: {self.fetcher.chapter_name}")
			else:
				print(f"fetching: ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}")
		while not self.fetcher.is_last_image():
			self._img_bin_list.append(requests.get(self.fetcher.image, headers=self._header).content)
			if not self._quiet:
				print(".", end="", flush=True)
			self.fetcher.next_image()
		self._img_bin_list.append(requests.get(self.fetcher.image, headers=self._header).content)
		if not self._quiet:
			print(".", end="", flush=True)

	def keep_full_chapter(self):
		"""
		Downloading all the images of the chapters and storing them where the output was specified.
		"""

		if not self._quiet:
			if isinstance(self.fetcher, StandaloneFetcher):
				print(f"downloading: {self.fetcher.chapter_name}")
			else:
				print(f"downloading: ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}")
		while not self.fetcher.is_last_image():
			if self._keep:
				self._img_bin_list.append(requests.get(self.fetcher.image, headers=self._header).content)
			self._dl_image()
			self.fetcher.next_image()

		if self._keep:
			self._img_bin_list.append(requests.get(self.fetcher.image, headers=self._header).content)
		self._dl_image()
		if not self._quiet and self._image:
			print("")

	def _skip(self):
		"""
		Skips the images as asked with the skip parameter.
		"""

		for loop in range(self._nskip):
			self.fetcher.next_image()

	def create_pdf(self):
		"""
		Creates the pdf at the output location with the fetched or the downloaded images of the current chapter.

		:raises EmptyChapter: the images of the current chapter were all blacklisted images and the pdf was empty
		"""
		print("Warning: the pdf creation engine will be changed for Wand in the next major release (3.0.0). Please do not forget to install imagemagick at that time", file=stderr)

		if not self._quiet:
			print("\nconverting...", end=" ")
		# loading the downloaded images if keep mode

		# removing the images found in the banlist
		self._img_bin_list = [img for img in self._img_bin_list if img not in self._banlist]

		if len(self._img_bin_list) > 0:
			# creating the pdf
			with Image() as pdf:
				for img_bin in self._img_bin_list:
					with contextlib.redirect_stderr(None):  # to mute alpha channel and ICC warnings as wand processes the image well anyway
						with Image(blob=img_bin) as img:
							pdf.sequence.append(img)
				pdf.save(filename=self._pdf_path)

			with open(self._pdf_path, "rb") as file:
				pdf = file.read()
			with open(self._pdf_path, "wb") as file:
				file.write(pdf.replace(b"/Producer (https://imagemagick.org)", b"/Producer (https://pypi.org/project/pyscandl/)")
						   .replace(
							b"/CreationDate", b"/Author <feff"+self.fetcher.author.encode("utf-16_be").hex().encode()+
											  b">\n/Keywords <feff"+self.fetcher.manga_name.encode("utf-16_be").hex().encode()+b">\n/CreationDate")
						   )  # manually adding the missing metadate from the pdf creation

			if not self._quiet:
				print("converted")
		else:
			raise EmptyChapter(self.fetcher.manga_name, self.fetcher.chapter_number)

	def go_to_chapter(self, chap_num):
		"""
		Make Pyscandl go to the asked chapter.

		:param chap_num: chapter number that was asked for
		:type chap_num: int/str/float
		"""

		# in case windows is the os, remove the banned characters
		if os.name == "nt":
			chapter_name = re_sub(r'[\\/*?:"<>|]', u"█", self.fetcher.chapter_name)
		else:
			chapter_name = self.fetcher.chapter_name

		self.fetcher.go_to_chapter(chap_num)
		self._path = f"{self._output}ch.{self.fetcher.chapter_number} {chapter_name}/"
		self._img_bin_list = []
		# prepares the next pdf path and name
		if self._tiny:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"
		else:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{self.fetcher.manga_name} - {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - {self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}{self.fetcher.manga_name} - ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"

	def next_chapter(self):
		"""
		Goes to the next chapter
		"""

		self.fetcher.next_chapter()

		# in case windows is the os, remove the banned characters
		if os.name == "nt":
			chapter_name = re_sub(r'[\\/*?:"<>|]', u"█", self.fetcher.chapter_name)
		else:
			chapter_name = self.fetcher.chapter_name

		self._path = f"{self._output}ch.{self.fetcher.chapter_number} {chapter_name}/"
		self._img_bin_list = []
		# prepares the next pdf path and name
		if self._tiny:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"
		else:
			if isinstance(self.fetcher, StandaloneFetcher):
				self._pdf_path = f"{self._output}{self.fetcher.manga_name} - {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - {self.fetcher.chapter_name}"
			else:
				self._pdf_path = f"{self._output}{self.fetcher.manga_name} - ch.{self.fetcher.chapter_number} {chapter_name}.pdf"
				self._name_metadata_pdf = f"{self.fetcher.manga_name} - ch.{self.fetcher.chapter_number} {self.fetcher.chapter_name}"

	def full_download(self):
		"""
		Does the full download process with what is specified when initializing the Pyscandl object
		"""

		try:
			# emulating a do while
			self._skip()
			counter = 1

			try:
				if self._keep or self._image:
					self.keep_full_chapter()
				else:
					self.full_chapter()

				if not self._image:
					try:
						self.create_pdf()
					except EmptyChapter:
						if not self._quiet:
							print("empty")
			except DelayedRelease as e:
				if not self._quiet:
					print(e)

			while not isinstance(self.fetcher, StandaloneFetcher) and not self.fetcher.is_last_chapter() and (self._all or counter < self._download_number or float(self.fetcher.chapter_number) < self._chapend):
				self.next_chapter()
				try:
					if self._keep or self._image:
						self.keep_full_chapter()
					else:
						self.full_chapter()

					if not self._image:
						try:
							self.create_pdf()
						except EmptyChapter:
							if not self._quiet:
								print("empty")
				except DelayedRelease as e:
					if not self._quiet:
						print(e)
				counter += 1
		except KeyboardInterrupt:
			if not self._quiet:
				print("\nmanual interruption")
		finally:
			self.fetcher.quit()
			if not self._quiet:
				print("end of the download")



import os
from platform import system
import sqlite3
import shutil

import cfscrape
from ..excepts import IsStandalone, FetcherNotFound, EmptyChapter, DelayedRelease, MangaNotFound
from ..Pyscandl import Pyscandl
from ..fetchers import FetcherEnum
from ..fetchers.fetcher import StandaloneFetcher


class Controller:
	__doc__ = """
	Object responsible of the autodl part of the program, all the logic related to it passes here
	"""

	def __init__(self, output:str=".", quiet:bool=False, tiny:bool=False):
		"""
		Initializes this instance of the autodl controller.
		If there is no sqlite database for autodl currently in existence a new one is created at:

		- ``~/.local/share/pyscandl/`` for linux
		- ``%APPDATA%/pyscandl/` for windows
		- ``~/Library/Preferences/pyscandl/`` for mac

		:param output: location where the outputted scans should be stored
		:type output: str
		:param quiet: should the program not output any information about what it is doing in the console
		:type quiet: bool
		:param tiny: should the name of every downloaded scan be minified and only include the chapter number and the chapter title
		:type tiny: bool
		"""

		# as it's in the users folder now it's OS dependent
		platform = system()
		if platform == "Linux":
			folder_path = os.path.expanduser("~/.local/share/pyscandl/")
		elif platform == "Windows":
			folder_path = os.path.expandvars("%APPDATA%/pyscandl/")
		elif platform == "Darwin":
			folder_path = os.path.expanduser("~/Library/Preferences/pyscandl/")
		else:
			raise OSError("The OS couldn't be detected, the db don't have a place to be stored")

		try:
			self._conn = sqlite3.connect(folder_path + "db.sqlite")
			self._curs = self._conn.cursor()
		except sqlite3.OperationalError as e:
			if str(e) == "unable to open database file":
				os.makedirs(folder_path)
				self._conn = sqlite3.connect(folder_path + "db.sqlite")
				self._curs = self._conn.cursor()
				self._curs.execute("""
				CREATE TABLE "manga" (
					"id" INTEGER PRIMARY KEY,
					"name" TEXT UNIQUE,
					"fetcher" TEXT,
					"link" TEXT,
					"archived" BOOL DEFAULT FALSE
				);
				""")

				self._curs.execute("""
				CREATE TABLE "chaplist" (
					"manga" INTEGER REFERENCES manga(id),
					"chapter" BLOB,
					CONSTRAINT unique_chap UNIQUE (manga, chapter)
				);
				""")
		self.output = output
		self.quiet = quiet
		self.tiny = tiny
		self.scrapper = cfscrape.create_scraper()
		self.missing_chaps = []
		self.downloads = 0

	def save(self):
		"""
		Saves the current state of the database in the ``db.sqlite`` file.
		"""

		self._conn.commit()

	def add(self, name:str, link:str, fetcher:str, chapters:list=None, archived=False):
		"""
		Adds a new scan entry to the ``db.sqlite`` file.

		:param name: name of the manga
		:type name: str
		:param link: link to the page of the manga *(same link that is used for the -l arg in other uses of pyscandl)*
		:type link: str
		:param fetcher: name of the associated fetcher
		:type fetcher: str
		:param chapters: list of the already possessed chapters that wont be downloaded again *(Optional)*
		:type chapters: list[int/float/str]
		:param archived: tell if the chapter is considered archived and if it'll be downloaded with autodl
		:type archived: bool

		:raises FetcherNotFound: the specified fetcher doesn't exist
		:raises IsStandalone: the specified fetcher is a standalone fetcher
		"""

		if chapters is None:
			chapters = []

		if fetcher.upper() not in [i.name for i in FetcherEnum]:
			raise FetcherNotFound(fetcher)
		if issubclass(FetcherEnum.get(fetcher), StandaloneFetcher):
			raise IsStandalone(name)

		self._curs.execute("""INSERT INTO manga("name", "fetcher", "link", "archived") VALUES (?, ?, ?, ?);""",
						   (
							   name,
							   fetcher.upper(),
							   link,
							   archived
						   ))
		self._curs.execute("""SELECT id FROM manga WHERE "name"=?""", (name,))
		manga_id = self._curs.fetchone()
		self._curs.executemany("""INSERT OR IGNORE INTO chaplist("manga", "chapter") VALUES (?, ?);""", [(manga_id[0], chap) for chap in chapters])

	def edit(self, name:str, link:str=None, fetcher=None, chapters:list=None, archived=None):
		"""
		Edits an already existing entry in the ``db.sqlite`` file.
		The :param name: is mandatory to find the correct entry and every other parameter specified will overwrite the existing values.

		:param name: name of the manga
		:type name: str
		:param link: link to the page of the manga *(same link that is used for the -l arg in other uses of pyscandl)*
		:type link: str
		:param fetcher: name of the associated fetcher
		:type fetcher: str
		:param chapters: list of the already possessed chapters that wont be downloaded again
		:type chapters: list[int/float/str]
		:param archived: tell if the chapter is considered archived and if it'll be downloaded with autodl
		:type archived: bool

		:raises IsStandalone: the specified fetcher is a standalone fetcher
		"""

		if link is not None:
			self._curs.execute("""UPDATE manga SET link=? WHERE name=?;""", (link, name))
		if fetcher is not None:
			standalone_check = FetcherEnum.get(fetcher)
			if issubclass(standalone_check, StandaloneFetcher):
				raise IsStandalone(name)
			self._curs.execute("""UPDATE manga SET fetcher=? WHERE name=?;""", (fetcher, name))
		if chapters is not None:
			self._curs.executemany("""INSERT OR IGNORE INTO chaplist("manga", "chapter") VALUES ((SELECT id FROM manga WHERE name=?), ?);""", [(name, chap) for chap in chapters])
		if archived is not None:
			self._curs.execute("""UPDATE manga SET archived=? WHERE name=?;""", (archived, name))

	def scan(self, name:str):
		"""
		Scans the asked manga for new and non downloaded chapters and adds them to the controller queue.

		:param name: name of the manga
		:type name: str

		:raise MangaNotFound: the asked manga isn't in the db
		"""

		self.missing_chaps.clear()
		manga = self._curs.execute("""SELECT * FROM manga WHERE name=?""", (name,)).fetchone()
		if manga is None:
			raise MangaNotFound(name)

		manga_fetcher = FetcherEnum.get(manga[2])
		manga_chapters = set(manga_fetcher.scan(link=manga[3]))
		possessed_chaps = {row[0] for row in self._curs.execute(
			"""SELECT chapter FROM chaplist WHERE manga=(SELECT id FROM manga WHERE name=?)""", (name,)
		).fetchall()}
		self.missing_chaps = list(manga_chapters - possessed_chaps)

		self.missing_chaps.sort()
		if not self.quiet:
			if self.missing_chaps:
				print(f"new chapter(s) for {name}: {', '.join(map(str, self.missing_chaps))}")
			else:
				print(f"no new chapter for {name}")

	def download(self, name:str, pdf:bool=True, keep:bool=False, image:bool=False):
		"""
		Start the download of the chapters of the asked manga that have their number in the scan results.

		:param name: name of the manga
		:type name: str
		:param pdf: tell if the result should be kept as a pdf
		:type pdf: bool
		:param keep: tell if the result should be kept as a pdf and as a collection of images
		:type keep: bool
		:param image: tell if the result should be kept as a collection of images
		:type image: bool
		"""

		manga = self._curs.execute("""SELECT * FROM manga WHERE name=?""", (name,)).fetchone()
		fetcher = FetcherEnum.get(manga[2])


		first = True
		chapter_id = 0
		while chapter_id < len(self.missing_chaps):
			try:
				if first:
					downloader = Pyscandl(fetcher, self.missing_chaps[chapter_id], self.output, link=manga[3], pdf=pdf, image=image, keep=keep, quiet=self.quiet, tiny=self.tiny)
					first = False
				else:
					downloader.go_to_chapter(self.missing_chaps[chapter_id])

				bad_image = True
				while bad_image:  # protect against bad downloads
					try:
						if keep or image:
							downloader.keep_full_chapter()
						elif pdf:
							downloader.full_chapter()
						if not image:
							downloader.create_pdf()
						bad_image = False
					except IOError:
						print(f"problem during download, retrying {name} chapter {self.missing_chaps[chapter_id]}")
						downloader.go_to_chapter(self.missing_chaps[chapter_id])

				self._curs.execute("""
					INSERT INTO chaplist (manga, chapter) VALUES ((SELECT id FROM manga where name=?), ?)
				""", (name, self.missing_chaps[chapter_id]))
				self.downloads += 1
			except EmptyChapter:
				if not self.quiet:
					print(f"skipping {name} chapter {self.missing_chaps[chapter_id]}: empty, wont be added in the downloaded list")
			except DelayedRelease as e:
				if not self.quiet:
					print(e)
			finally:
				chapter_id += 1

		if self.missing_chaps:
			try:
				downloader.fetcher.quit()
			except UnboundLocalError:
				pass  # only available chapter was a delayed release or an empty chapter

		# remove the directory if there is no chapter
		try:
			folders = list(os.walk(self.output))[1:]
			for folder in folders:
				if not folder[2]:
					os.rmdir(folder[0])
		except OSError:
			pass

	def list_mangas(self, all=False, only=False, format=True):
		"""
		Gives the list of all the names of the mangas in the ``db.sqlite`` file. if the db is empty, returns None

		:param all: get also the archived mangas
		:type all: bool
		:param only: get only the archived mangas
		:type only: bool

		:rtype: list
		"""

		titles = []

		if all:
			manga = self._curs.execute("""SELECT * FROM manga ORDER BY name""").fetchall()
		elif only:
			manga = self._curs.execute("""SELECT * FROM manga WHERE archived=true ORDER BY name""").fetchall()
		else:
			manga = self._curs.execute("""SELECT * FROM manga WHERE archived=false ORDER BY name""").fetchall()
		if not format:
			return [row[1] for row in manga]

		max_len_title = len(max(manga, key=lambda x: len(x[1]))[1])
		fetchers = [row[2] for row in manga]
		max_len_fetcher = len(max(fetchers, key=lambda x: len(x)))
		for id, title, fetcher, link, archived in manga:
			if archived and (all or only):
				titles.append("{title:<{title_size}}|{fetcher:^{fetcher_size}}|{}".format(
					"    Archived",
					title=title,
					title_size=max_len_title + 4,
					fetcher=fetcher,
					fetcher_size=max_len_fetcher + 8
				))
			elif not archived and not only:
				titles.append("{title:<{title_size}}|{fetcher:^{fetcher_size}}|".format(
					title=title,
					title_size=max_len_title + 4,
					fetcher=fetcher,
					fetcher_size=max_len_fetcher + 8
				))
		return titles

	def manga_info(self, name):
		"""
		Fet the infos about a specific manga.

		:param name: name of the manga
		:type name: str

		:rtype: tuple
		:returns: name, fetcher, link, archived, list[chapters]
		"""

		try:
			info = self._curs.execute("""SELECT * FROM manga WHERE name=?""", (name,)).fetchone()
			chaps = self._curs.execute("""SELECT chapter FROM chaplist WHERE manga=? ORDER BY chapter DESC""", (info[0],)).fetchall()
			return (*info[1:], [row[0] for row in chaps])
		except TypeError:
			return None

	def delete_manga(self, name):
		"""
		Deletes a manga from the ``db.sqlite`` file.

		:param name: name of the manga

		:return: confirms the deletion
		:rtype: bool
		"""
		if self._curs.execute("""SELECT * FROM manga where name=?;""", (name,)).fetchone():
			self._curs.execute("""DELETE from chaplist where manga=(SELECT id FROM manga where name=?);""", (name,))
			self._curs.execute("""DELETE FROM manga where name=?;""", (name,))
			return True
		else:
			return False

	def rm_chaps(self, name, *rm_chaps):
		"""
		Remove the listed chapters from the asked manga

		:param name: name of he manga
		:type name: str
		:param rm_chaps: list of all the chapters that have to be removed
		:type rm_chaps: str

		:return: confirms the deletion
		:rtype: bool
		"""

		vals = [(float(chap) if "." in chap else int(chap), name) for chap in rm_chaps[0]]
		return bool(self._curs.executemany("""DELETE FROM chaplist where chapter=? and manga=(SELECT id FROM manga WHERE name=?)""", vals).rowcount)

	def db_import(self, path:str):
		"""
		Takes an external sqlite file path and put its content as the new database for pyscandl.

		:param path: path to the .sqlite file to import as the database for autodl and manga
		:type path: str
		"""

		# as it's in the users folder now it's OS dependent
		platform = system()
		if platform == "Linux":
			folder_path = os.path.expanduser("~/.local/share/pyscandl/")
		elif platform == "Windows":
			folder_path = os.path.expandvars("%APPDATA%/pyscandl/")
		elif platform == "Darwin":
			folder_path = os.path.expanduser("~\Library\Preferences/pyscandl/")
		else:
			raise OSError("The OS couldn't be detected, the db don't have a place to be stored")

		shutil.copy(path, folder_path + "db.sqlite")

	def db_export(self, path:str):
		"""
		Saves a copy of the current database to the folder specified.

		:param path: path to the save folder location
		:type path: str
		"""

		# as it's in the users folder now it's OS dependent
		platform = system()
		if platform == "Linux":
			folder_path = os.path.expanduser("~/.local/share/pyscandl/")
		elif platform == "Windows":
			folder_path = os.path.expandvars("%APPDATA%/pyscandl/")
		elif platform == "Darwin":
			folder_path = os.path.expanduser("~\Library\Preferences/pyscandl/")
		else:
			raise OSError("The OS couldn't be detected, the db don't have a place to be stored")

		if not os.path.exists(path):
			os.makedirs(path)
		shutil.copy(folder_path + "db.sqlite", path)



from sys import stderr, modules
from platform import system
from os import path, makedirs, remove
import sqlite3
import json

from pyscandl.modules import arg_parser, Pyscandl
from pyscandl.modules.fetchers import FetcherEnum
from pyscandl.modules.autodl import Controller
from pyscandl.modules.excepts import DownedSite, MangaNotFound
import xml.etree.ElementTree


def main():
	args = arg_parser.get_parser().parse_args()

	if args.subparser == "gui":
		main_gui()
	elif args.subparser == "manual":
		fetcher = FetcherEnum.get(args.fetcher)
		pyscandl = Pyscandl(fetcher, chapstart=args.chapter_start, output=args.output, pdf=args.pdf,
									 keep=args.keep, image=args.image, all=args.all, link=args.link, manga=args.manga,
									 download_number=args.download_number, chapend=args.chapter_end, quiet=args.quiet,
									 skip=args.skip, tiny=args.tiny)
		pyscandl.full_download()

	elif args.subparser == "manga":
		print("Warning: the current db will be replaced wy a new system in the next major release (3.0.0). Please do not forget the migration at that time", file=stderr)

		if args.list or args.list_all or args.list_only:
			ml = Controller().list_mangas(all=args.list_all, only=args.list_only)
			if ml:
				list = "\n- " + "\n- ".join(ml)
				print(f"current mangas in the autodl db are:{list}")
			else:
				print("there are currently no mangas in autodl, you may consider adding some to it with manga add")

		elif args.import_db:
			controller = Controller()
			controller.db_import(args.import_db)
			controller.save()

		elif args.export_db:
			Controller().db_export(args.export_db)

		elif args.manga_subparser == "scan":
			try:
				infos = Controller()
				if args.name:
					try:
						infos.scan(args.name)
					except MangaNotFound as e:
						print(e)
				else:
					mangas = [row[0] for row in infos._curs.execute("""SELECT name FROM manga WHERE archived=false ORDER BY name""").fetchall()]
					for manga in mangas:
						infos.scan(manga)
			except KeyboardInterrupt:
				print("\nmanual interruption")

		elif args.manga_subparser == "info":
			infos = Controller().manga_info(args.name)
			if infos is None:
				print(f"manga '{args.name}' not in the list, you may consider adding it to it with manga add")
			elif infos[4]:
				print(f"{args.name}:\n",
					  f"\tmanga link: {infos[2]}\n",
					  f"\tfetcher: {infos[1]}\n",
					  f"\tnumber of chapters already downloaded: {len(infos[4])}\n",
					  f"\tlast chapter downloaded: {infos[4][0]}\n",
					  f"\tarchived: {bool(infos[3])}")
			else:
				print(f"{args.name}:\n",
					  f"\tmanga link: {infos[2]}\n",
					  f"\tfetcher: {infos[1]}\n",
					  f"\tno chapter downloaded yet\n",
					  f"\tarchived: {bool(infos[3])}")

		elif args.manga_subparser == "add":
			controller = Controller()
			if args.chap:
				chaps = [float(chap) if "." in chap else int(chap) for chap in args.chap]
			else:
				chaps = []
			controller.add(args.name, args.link, args.fetcher, chaps, args.archived)
			controller.save()

		elif args.manga_subparser == "edit":
			controller = Controller()
			if args.chap:
				chaps = [float(chap) if "." in chap else int(chap) for chap in args.chap]
			else:
				chaps = None

			if args.archive:
				archive = True
			elif args.unarchive:
				archive = False
			else:
				archive = None

			controller.edit(args.name, args.link, args.fetcher, chaps, archive)
			controller.save()

		elif args.manga_subparser == "chaplist":
			info = Controller().manga_info(args.name)
			if info is not None:
				chaps = info[4]
				if chaps:
					if not args.quiet:
						print(f"the already downloaded chapters for '{args.name}' are:", end=" ")
					print(' '.join([str(chap) for chap in chaps]))
				else:
					print(f"no chapter downloaded yet for '{args.name}'")
			elif not args.quiet:
				print(f"manga '{args.name}' not in the list, you may consider adding it to it with manga add")

		elif args.manga_subparser == "rmchaps":
			controller = Controller()
			if controller.rm_chaps(args.name, args.chap):
				controller.save()
				if not args.quiet:
					print(f"deletion of the chapters {', '.join(args.chap)} from {args.name} sucessfull")
			else:
				if not args.quiet:
					print(f"no chapters removed for {args.name}")

		elif args.manga_subparser == "delete":
			controller = Controller()
			if controller.delete_manga(args.name):
				controller.save()
				if not args.quiet:
					print(f"deletion of {args.name} successful")
			else:
				if not args.quiet:
					print(f"manga {args.name} not found")

		elif args.manga_subparser == "migrate":
			if not args.quiet:
				print("WARNING: if there is already a database that was migrated before it will be erased in the process !")
			if args.quiet or input("Continue ? [y/N]").lower() == "y":
				if not args.quiet:
					print("Creating new db file...", end=" ")

				# as it's in the users folder now it's OS dependent
				platform = system()
				if platform == "Linux":
					folder_path = path.expanduser("~/.local/share/pyscandl/")
					# removing the old db if it exists
					try:
						remove(folder_path+"db.sqlite")
					except FileNotFoundError:
						pass

					# creating the new db
					makedirs(folder_path, exist_ok=True)
					conn = sqlite3.connect(folder_path + "db.sqlite")
				elif platform == "Windows":
					folder_path = path.expandvars("%APPDATA%/pyscandl/")
					# removing the old db if it exists
					try:
						remove(folder_path+"db.sqlite")
					except FileNotFoundError:
						pass

					# creating the new db
					makedirs(folder_path, exist_ok=True)
					conn = sqlite3.connect(folder_path + "db.sqlite")
				elif platform == "Darwin":
					folder_path = path.expanduser("~\Library\Preferences/pyscandl/")
					# removing the old db if it exists
					try:
						remove(folder_path+"db.sqlite")
					except FileNotFoundError:
						pass

					# creating the new db
					makedirs(folder_path, exist_ok=True)
					conn = sqlite3.connect(folder_path + "db.sqlite")
				else:
					raise OSError("The OS couldn't be detected, the db don't have a place to be stored")
				curs = conn.cursor()

				if not args.quiet:
					print("Loading the old db file...")
				try:
					with open(f"{path.dirname(modules['pyscandl.modules.autodl'].__file__)}/db.json", "r") as data:
						old_db = json.load(data)
				except FileNotFoundError:
					old_db = {}

				if not args.quiet:
					print("Creating new tables...", end=" ")
				curs.execute("""
				CREATE TABLE IF NOT EXISTS "manga" (
					"id" INTEGER PRIMARY KEY,
					"name" TEXT UNIQUE,
					"fetcher" TEXT,
					"link" TEXT,
					"archived" BOOL DEFAULT FALSE
				);
				""")

				curs.execute("""
				CREATE TABLE IF NOT EXISTS "chaplist" (
					"manga" INTEGER REFERENCES manga(id),
					"chapter" BLOB,
					CONSTRAINT unique_chap UNIQUE (manga, chapter)
				);
				""")

				if not args.quiet:
					print("Transfering data...")
				for key, value in old_db.items():
					if not args.quiet:
						print(f"{key}: autodl data...", end=" ")
					curs.execute("""INSERT INTO manga("name", "fetcher", "link", "archived") VALUES (?, ?, ?, ?);""",
								 (
									 key,
									 value.get("fetcher").upper(),
									 value.get("link"),
									 value.get("archived")
								 ))
					if not args.quiet:
						print("already downloaded chapters...")
					curs.execute("""SELECT id FROM manga WHERE "name"=?""", (key,))
					manga_id = curs.fetchone()
					curs.executemany("""INSERT OR IGNORE INTO chaplist("manga", "chapter") VALUES (?, ?);""", [(manga_id[0], chap) for chap in value.get("chapters")])

				conn.commit()
				conn.close()
			else:
				print("Cancelling migration")

	elif args.subparser == "autodl":
		print("Warning: the current db will be replaced wy a new system in the next major release (3.0.0). Please do not forget the migration at that time", file=stderr)

		autodl = Controller(args.output, args.quiet, args.tiny)
		# to be sure to save progress done in case of interruption
		try:
			for name in [row[0] for row in autodl._curs.execute("SELECT name FROM manga WHERE archived=false ORDER BY name").fetchall()]:
				# currently having problems with the xml tree fetching sometimes so giving a retry possibility to happen
				tries_left = 3
				while tries_left > 0:
					try:
						autodl.scan(name)
						success = True
						tries_left = 0
					except xml.etree.ElementTree.ParseError:
						if not args.quiet:
							print(f"problem with the fetching for {name}, retrying...")
						success = False
						tries_left -= 1
					except DownedSite:
						# the website can't be accessed for the time being so no retrying
						success = False
						tries_left = 0
				if success:
					try:
						autodl.download(name, pdf=args.pdf, keep=args.keep, image=args.image)
					except DownedSite:
						print(f"can't access {name}, please retry it later")
				elif not args.quiet:
					print(f"can't access {name}, please retry it later")
		except KeyboardInterrupt:
			if not args.quiet:
				print("\nmanual interruption")
		finally:
			autodl.save()
			if not args.quiet:
				print(f"{autodl.downloads} chapters downloaded")

def main_gui():
	from .gui import pyscandl_qt


if __name__ == "__main__":
	main()



from PyQt5.QtCore import QRunnable, pyqtSlot, QObject, pyqtSignal
import traceback, sys


class WorkerSignals(QObject):
    """
    Defines the signals available from a running worker thread.

    Supported signals are:

    finished
        No data

    error
        tuple (exctype, value, traceback.format_exc() )

    result
        object data returned from processing, anything

    """
    finished = pyqtSignal(bool)
    error = pyqtSignal(tuple)
    result = pyqtSignal(object)


class Worker(QRunnable):
    def __init__(self, fn, *args, **kwargs):
        """
        Worker thread

        Inherits from QRunnable to handler worker thread setup, signals and wrap-up.

        :param callback: The function callback to run on this worker thread. Supplied args and
                         kwargs will be passed through to the runner.
        :type callback: function
        :param args: Arguments to pass to the callback function
        :param kwargs: Keywords to pass to the callback function
        """

        super(Worker, self).__init__()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

    @pyqtSlot()
    def run(self):
        """
        Initialise the runner function with passed args, kwargs.
        """

        try:
            self.fn(*self.args, **self.kwargs)
        except:
            traceback.print_exc()
            exctype, value = sys.exc_info()[:2]
            self.signals.error.emit((exctype, value, traceback.format_exc()))
        finally:
            self.signals.finished.emit(True)



import re
import secrets
import requests
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import ParseError
from bs4 import BeautifulSoup

from ..excepts import MangaNotFound, EmptyChapter, DownedSite
from .fetcher import Fetcher


class Fanfox(Fetcher):
    __doc__ = """
    This is the fetcher in charge of https://fanfox.net/ 
    """

    def __init__(self, link:str=None, manga:str=None, chapstart=1):
        """
        Initializes the instance of the fanfox fetcher, it needs either manga or link to work.

        :param link: link of the scan wanted
        :type link: str
        :param manga: manga name with all the non alpha numeric characters with "_", ex: fullmetal_alchemist
        :type manga: str
        :param chapstart: number of the chapter that the download is supposed to start
        :type chapstart: int/float/str

        :raises MangaNotFound: the scan asked for can't be found
        """

        super().__init__(link, manga, chapstart)
        self._header = {"User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/77.0.3865.120 Safari/537.36",
                        "Referer": "test"}

        self.domain = ".fanfox.net"
        # creating the chapter link
        if link is not None:
            if link[-1] == "/":
                self._link = link
            else:
                self._link = link + "/"
            self.manga_name = self._link.split("/")[-2].replace("_", " ").title()
        else:
            self._link = f"https://fanfox.net/manga/{manga.replace(' ', '_').lower()}/"
            self.manga_name = manga.replace("_", " ").title().replace("/", "_")

        req = requests.get(self._link)
        soup = BeautifulSoup(req.text, "html.parser")
        if req.url != self._link:
            raise MangaNotFound(self.manga_name)
        try:
            self.author = soup.find("p", class_="detail-info-right-say").findChild("a").text
        except AttributeError:
            self.author = "unknown"

        # regex i can't get rid of... :/
        self._c_eval_args = re.compile(r"eval\(function\(p,a,c,k,e,d\).+?p;}\('(?P<pattern>.+?)',(?P<counter1>\d+?),(?P<counter2>\d+?),'(?P<str_table>.+?)'\.split")
        self._c_decode_word_replacer = re.compile(r"\b\w+\b")

        self.go_to_chapter(chapstart)

    def _decode(self, obfuscated:str):
        """
        Method used to de-obfuscate the web-pages of the scan to get the image info

        :param obfuscated: html code of the web-page of the scan
        :type obfuscated: str

        :return decoded: de-obfuscated image info that was contained in the page
        :rtype decoded: str
        """

        match = self._c_eval_args.search(obfuscated)
        pattern = match.group("pattern").replace("\\", "")
        counter1 = int(match.group("counter1"))
        counter2 = int(match.group("counter2"))
        str_table = match.group("str_table").split("|")
        pattern_dict = {}

        def calc_dict_indices(iter_number):
            if iter_number < counter1:
                prefix = ""
            else:
                prefix = calc_dict_indices(int(iter_number / counter1))

            remainder = iter_number % counter1

            if remainder > 35:
                return prefix + chr(remainder + 29)
            else:
                return prefix + "0123456789abcdefghijklmnopqrstuvwxyz"[remainder]

        for i in range(counter2 - 1, -1, -1):
            dict_index = calc_dict_indices(i)
            pattern_dict[dict_index] = str_table[i] if str_table[i] else dict_index

        return self._c_decode_word_replacer.sub(repl=lambda word: pattern_dict[word.group()], string=pattern)

    def go_to_chapter(self, chap):
        """
        Go to the specified chapter.

        :param chap: number of the specified chapter

        :raises MangaNotFound: the asked chapter doesn't exist
        :raises EmptyChapter: the chapter exists but is void of images
        """

        self._image_list = []
        self.npage = 1
        self.chapter_number = str(chap).split(".")[0].zfill(3)
        if "." in str(chap):
            self.chapter_number += "." + str(chap).split(".")[1]

        url = f"{self._link}c{self.chapter_number}/1.html"
        self._req = requests.get(url, cookies={"isAdult": "1"})
        if "<title>404</title>" in self._req.text:
            raise MangaNotFound(f"{self.manga_name}, chapter {self.chapter_number}")
        if '<p class="detail-block-content">No Images</p>' in self._req.text:
            raise EmptyChapter(self.manga_name, self.chapter_number)
        self._soup = BeautifulSoup(self._req.text, "html.parser")

        self._mono = self._req.text.count("dm5_key") == 1
        self.chapter_name = " ".join(self._soup.find("p", class_="reader-header-title-2").text.split("Ch.")[-1].split(" ")[1:])

        if self._mono:
            self._mono_go_to_chap()
        else:
            self._multi_go_to_chap()

        self.image = self._image_list[0]

        if requests.get(self.image).status_code == 404:  # if a chapter has TBE as a volume it doesn't work without stating vTBE to redo but with it
            self._image_list = []
            self.npage = 1
            self.chapter_number = str(chap).split(".")[0].zfill(3)
            if "." in str(chap):
                self.chapter_number += "." + str(chap).split(".")[1]

            url = f"{self._link}vTBE/c{self.chapter_number}/1.html"
            self._req = requests.get(url, cookies={"isAdult": "1"})
            if "<title>404</title>" in self._req.text:
                raise MangaNotFound(f"{self.manga_name}, chapter {self.chapter_number}")
            if '<p class="detail-block-content">No Images</p>' in self._req.text:
                raise EmptyChapter(self.manga_name, self.chapter_number)
            self._soup = BeautifulSoup(self._req.text, "html.parser")

            self._mono = self._req.text.count("dm5_key") == 1
            self.chapter_name = " ".join(
                self._soup.find("p", class_="reader-header-title-2").text.split("Ch.")[-1].split(" ")[1:])

            if self._mono:
                self._mono_go_to_chap()
            else:
                self._multi_go_to_chap()

            self.image = self._image_list[0]

    def _mono_go_to_chap(self):
        """
        Method used to get the obfuscated image info in mono type scan chapters.
        *mono type chapters are characterized by having all the images displayed one after another in a single page*
        """

        out = self._decode(self._req.text)
        self._image_list = [f"https:{page.split('?')[0]}" for page in out.split("=[")[1].split("];")[0].replace("'", "").split(",")]  # get the inside of the js list of images and transforms it onto the proper python list of image
        self._last_page = len(self._image_list)

    def _multi_go_to_chap(self):
        """
        Method used to get obfuscated image info in multi type scan chapters.
        *multi type chapters are characterized by having one image per page*
        """

        cid = self._req.text.split("var chapterid =")[1].split(";")[0]
        self._last_page = max([int(i.text) for i in self._soup.find_all(lambda x: x.has_attr("data-page") and x.text.isdigit())])

        for page in range(1, self._last_page+1):
            if page > len(self._image_list):
                req2 = requests.get(f"https://fanfox.net/chapterfun.ashx?cid={cid}&page={page}&key={secrets.token_hex(8)}", headers=self._header)
                out = self._decode(req2.text)
                root = out.split("(){")[1].split('pix="')[1].split('";')[0]
                chaps = [i.split("?")[0] for i in out.split("pvalue=[")[1].split("];")[0].split('"') if i != "" and i != ","]

                for i in chaps:
                    if i[0] != "/":
                        i = "/" + i
                    self._image_list.append(f"https:{root + i}")

    def next_image(self):
        self.image = self._image_list[self.npage]
        self.npage += 1
        self.ext = self.image.split(".")[-1]

    def is_last_image(self):
        # automatically ignoring the last page of add
        return self.npage == len(self._image_list)-1

    def next_chapter(self):
        self.go_to_chapter(self._soup.find("a", class_="reader-main-next").get("href").split("/")[-2][1:])

    def is_last_chapter(self):
        try:
            self._soup.find("a", class_="reader-main-next").get("href").split("/")[-2][1:]
            return False
        except AttributeError:
            return True

    @classmethod
    def scan(cls, link:str=None, manga:str=None):
        if link is not None:
            manga_id = link.split("manga/")[1].split("/")[0]
        elif manga is not None:
            manga_id = manga.replace(" ", "_")

        rss_url = f"https://fanfox.net/rss/{manga_id}.xml"

        try:
            req = requests.get(rss_url)
            if req.status_code in (522, 504):
                raise DownedSite("https://fanfox.net")
            xml_root = ET.fromstring(req.text)
        except ParseError:
            raise MangaNotFound(manga_id)

        chap_list = []

        for item in xml_root.iter("item"):
            chap_nb = item.find("link").text.split("/")[-2][1:]
            if "." in chap_nb:
                chap_nb = float(chap_nb)
            else:
                chap_nb = int(chap_nb)

            chap_list.append(chap_nb)

        return list(set(chap_list))





import setuptools
from setuptools import setup
from pyscandl import __version__

with open("README.rst", "r") as f:
    readme = f.read()

requires = [
    "requests~=2.25.1",
    "wand~=0.6.5",
    "Pillow~=8.1.0",
    "cfscrape~=2.1.1",
    "beautifulsoup4~=4.9.3",
    "PyQt5~=5.15.2",
]

setup(
    name='pyscandl',
    version=__version__,
    packages=setuptools.find_packages(),
    url='https://pyscandl.readthedocs.io/',
    license='BSD 3-Clause License',
    author='Thomas MONTERO | Ara0n',
    author_email='thomas99.montero@gmail.com',
    description='a scan downloader in python',
    keywords="scraping web scan download manga comics mangadex webtoon nhentai fanfox naver frscan english french cli gui qt",
    project_urls={
        "Documentation": "https://pyscandl.readthedocs.io",
        "Source Code": "https://github.com/Ara0n/pyscandl",
        "Bug Tracker": "https://github.com/Ara0n/pyscandl/issues",
    },
    python_requires=">=3.7",
    entry_points={"console_scripts": ["pyscandl = pyscandl.main:main", "pyscandl-qt = pyscandl.main:main_gui"]},
    long_description=readme,
    long_description_content_type='text/x-rst',
    install_requires=requires,
    package_data={
        "pyscandl": ["banlist/*"],
        "": ["README.rst", "CHANGELOG.rst"]
    }
)
